#ifndef NEER_ERROR_H
#define NEER_ERROR_H

#include <stdexcept>
#include <string>
#include <memory>

#define ERR(fmt,...) Error(nerr::sfmt("%s:%d: " fmt, __FILE__, __LINE__, ##__VA_ARGS__))

namespace nerr
{
    template <typename... Args>
    std::string sfmt(const std::string &format, Args... args)
    {
        int size_s = std::snprintf(nullptr, 0, format.c_str(), args...) + 1; // Extra space for '\0'
        if (size_s <= 0)
        {
            throw std::runtime_error("Error during formatting string.");
        }
        auto size = static_cast<size_t>(size_s);
        std::unique_ptr<char[]> buf(new char[size]);
        std::snprintf(buf.get(), size, format.c_str(), args...);
        return std::string(buf.get(), buf.get() + size - 1); // We don't want the '\0' inside
    }

    class Error : public std::runtime_error
    {
    public:
        Error(const std::string &s) : runtime_error(s){};
        Error(const char *s) : runtime_error(s){};
        Error(): runtime_error(""){};

        inline friend bool operator==(const Error& l, const Error& r)
        {
            return std::string(l.what()) == std::string(r.what());
        }

        inline Error operator+(const Error& other)
        {
            return Error(std::string(this->what()) + "\n" + other.what());
        }
    };
}

#endif