#ifndef NEER_OPTION_H
#define NEER_OPTION_H

#define None(T) (nerr::Option<T>())
#define Some(v) (nerr::Option(v))

namespace nerr
{
    template <typename T, typename E> class Result;

    template <typename T>
    class Option
    {

    public:
        Option(T obj) : m_some(obj), m_notnull(true){};
        Option() : m_notnull(false){};

        Option<T>& operator=(const T& value) {insert(value); return *this;};
        inline friend bool operator==(const Option<T>& l, const Option<T>& r) {
            if(!l.m_notnull && !r.m_notnull)
            {
                return true;
            }
            if(l.m_notnull && r.m_notnull)
            {
                return l.m_some == r.m_some;
            }
            return false;
        };

        inline bool is_none(){return !m_notnull;};
        inline bool is_some(){return m_notnull;};
        inline T& unwrap()
        {
            if(m_notnull)
            {
                return m_some;
            }
            throw Error("Object is None");
        }
        inline T& get_or_insert(T value)
        {
            if(!m_notnull)
            {
                insert(value);
            }
            return m_some;
        }
        inline T& insert(T value)
        {
            m_some = value;
            m_notnull = true;
            return m_some;
        }

        template <typename E>
        inline Result<T, E> ok_or(E err)
        {
            if(m_notnull)
            {
                return Result<T, E>(m_some);
            }
            return Result<T, E>(err);
        }

        template<typename U>
        inline Option<U> map(const std::function<U(T)>& fn)
        {
            if(is_some())
            {
                return fn(m_some);
            }
            return Option<U>();
        }

        template<typename U>
        inline U map_or(U defv, const std::function<U(T)>& fn)
        {
            if(is_some())
            {
                return fn(m_some);
            }
            return defv;
        }

    private:
        T m_some;
        bool m_notnull;
    };
}
#endif